#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <termios.h>
#include <unistd.h>
#include "phone.h"
#include "language.h"

/*
 * Version information: <VMAJOR>.<VMINOR>.<VCOUNT>
 */
#define VMAJOR		0
#define VMINOR		3
#define VCOUNT		0

#define TOKEN_BASE	10000

#define SUBJECT		10000
#define TYPE		10001
#define FROM		10002
#define STARTDATE	10003
#define ENDDATE		10004
#define PUBLICACCESS	10005
#define PRINCIPAL	10006
#define STARTTIME	10007
#define ENDTIME		10008
#define STARTDATETIME	10009
#define ENDDATETIME	10010
#define ALARM		10011
#define LOCATION	10012
#define CATEGORIES	10013
#define ROOM		10014
#define TOPIC		10015
#define SENDTO		10016
#define REVISIONS	10017
#define INETFROM	10018
#define INETSENDTO	10019
#define INETCOPYTO	10020
#define INETREQUIREDNAMES	10021
#define MESSAGEID	10022
#define APPTUNID	10023
#define POSTEDDATE	10024
#define SEQUENCENUM	10025
#define BUSYPRIORITY	10026
#define REQUIREDATTENDEES	10027
#define ALARMOFFSET	10028
#define APPOINTMENTTYPE	10029
#define REPEATS		10030
#define ALARMS		10031
#define STARTTIMEZONE	10032
#define ENDTIMEZONE	10033

#define COMMENT		11000
#define false		0
#define true		1

struct _slot
{
	time_t start;
	time_t end;
	char *summary;
	struct _slot *next;
};

struct _X400
{
	char *cn;
	char *ou;
	char *o;
	char *c;
	char *inet;
};

struct _vcalendar_datetime
{
	int day;
	int mon;
	int year;
	int hour;
	int min;
	int sec;
};

struct _vcalendar_chaindt
{
	int day;
	int mon;
	int year;
	int hour;
	int min;
	int sec;
	struct _vcalendar_chaindt *next;
};

struct _vcalendar_head
{
	char prodid[81];
	char version[11];
};

struct _vcalendar_attendee
{
	char *cn;
	int rsvpt;
	char partstat[31];
	char role[31];
	char *mailto;
	struct _vcalendar_attendee *next;
};

struct _vcalendar_trigger
{
	char *value;
	char *duration;
};

struct _vcalendar_valarm
{
	char action[31];
	struct _vcalendar_trigger *vcal_trigger;
};

// Z=-1$DO=1$DL=3 -1 1 10 -1 1$ZX=70$ZN=W. Europe
struct _vcalendar_timezone
{
	int base;		// Z= Differenz zu GMT
	int doffset;		// DO= Sommerzeit?
	char *dlocal;		// DL= ??
	int zx;			// ZX= Interner Region Code?
	char *zn;		// ZN= Region in Klartext
};

struct _vcalendar_body
{
	struct _vcalendar_head *vcal_head;
	struct _vcalendar_datetime *created;
	char *UID;
	int sequence;
	struct _vcalendar_datetime *last_modified;
	struct _vcalendar_datetime *dtstamp;
	char *organizer;
	struct _vcalendar_attendee *vcal_att;
	char *description;
	char *summary;
	char class[31];
	int priority;
	char *categories;
	int x_pilotid;
	int x_pilotstat;
	char *duration;
	struct _vcalendar_datetime *dtstart;
	struct _vcalendar_datetime *dtend;
	struct _vcalendar_valarm *vcal_valarm;
	int chstart_anz;
	struct _vcalendar_chaindt *chstart;
	int chend_anz;
	struct _vcalendar_chaindt *chend;
	struct _vcalendar_datetime *dstart;
	struct _vcalendar_datetime *dend;
	struct _vcalendar_timezone *tzstart;
	struct _vcalendar_timezone *tzend;
	char *type;
	int alarm;
	int alarm_offset;
	int appointment_type;
	int repeats;
	char *location;
};

struct _vcalendar_body vcal;
struct _vcalendar_head vcal_head;
struct _slot *slot;

char Keys[35][30] = { "Subject", "Type", "From", "StartDate",
		      "EndDate", "$PublicAccess", "Principal",
		      "StartTime", "EndTime", "StartDateTime",
		      "EndDateTime", "$Alarm", "Location", "Categories",
		      "ROOM", "TOPIC", "SendTo", "$Revisions", "INetFrom",
		      "INetSendTo", "INetCopyTo", "INetRequiredNames",
		      "$MessageID", "APPTUNID", "PostedDate",
		      "SEQUENCENUM", "$BusyPriority",
		      "REQUIREDATTENDEES", "$AlarmOffset",
		      "AppointmentType", "Repeats", "Alarms",
		      "StartTimeZone", "EndTimeZone", "\0\0" };

/*
 * globale Variablen fuer die Haltung der Kommandozeilenparameter.
 */
char Kalendar[512];
int ZeitDifferenz;
char Outfile[512];
char SerialDevice[512];
speed_t Baudrate;
int Handy;
int Organizer;
int LocalTime;
int Verbose;

int rfrom;
int rto;
static int SlotNum = 1;
static int wVcalNum = 0;

char *WorkComment (char *line, char *rest);
void WorkPostedDate (char *line, char *rest);
void WorkAPPTUNID (char *line, char *rest);
void WorkINetSendTo (char *line, char *rest);
void WorkSequenceNum (char *line, char *rest);
void WorkINetRequiredNames (char *line, char *rest);
void WorkSubject (char *line, char *rest);
void WorkPublicAccess (char *line, char *rest);
void WorkBusyPriority (char *line, char *rest);
void WorkCategories (char *line, char *rest);
void WorkStartTime (char *line, char *rest);
void WorkStartDateTime (char *line, char *rest);
void WorkEndTime (char *line, char *rest);
void WorkEndDateTime (char *line, char *rest);
void WorkType (char *line, char *rest);
void WorkAlarm (char *line, char *rest);
void WorkRepeats (char *line, char *rest);
void WorkAlarmOffset (char *line, char *rest);
void WorkAppointmentType (char *line, char *rest);
void WorkLocation (char *line, char *rest);
void WorkStartDate (char *line, char *rest);
void WorkEndDate (char *line, char *rest);
void WorkStartTimeZone (char *line, char *rest);
void WorkEndTimeZone (char *line, char *rest);
void WriteVEvent (FILE *fp);
void WriteHandy (int fd);
void FreeVEvent (void);
struct _vcalendar_datetime *ConvertDateTime (char *ret, char *str);
int TimeZone (struct _vcalendar_datetime *date, int zeit);
struct _vcalendar_timezone *GetTimeZone (char *zone);
int AddSlot (time_t start, time_t end, char *summary);
struct _slot *FindSlot (time_t start, time_t end, char *summary);
struct _slot *FindSlotSummary (char *summary);
void FreeSlot ();
char *GetNotesAddress (char *str);
struct _X400 *GetX400Address (char *str);
char *WinToUnix (char *str);
char *UnixToAscii (char *str);
char *GetLine (char *pos);
char *AvoidNull (char *str, int len);
char *MyStrdup (char *str);
char *ltrim (char *str);
char *rtrim (char *str);
void ShowHelp (void);
void ShowVersion (void);

//
// Main function. Program starts here.
//
int main (int argc, char *argv[])
{
int fd, fdh, i, zeile;
FILE *fp;
char *buffer, *pos, *p, *dot;
char hv0[512], hv1[81];
struct stat st;

	LocalTime = ZeitDifferenz = 0;
	Outfile[0] = 0;
	Kalendar[0] = 0;
	Baudrate = 0;
	Handy = 0;
	Organizer = 0;
	Verbose = 0;
	slot = NULL;
	memset (&vcal, 0, sizeof(vcal));

	ShowVersion ();
	InitLang ();

	/*
	 * Ueberpruefen der Argumente, welche ueber die Kommandozeile
	 * uebergeben wurden.
	 */
	if (argc < 2)
	{
	   lfprintf (10000, stderr, "No parameters found!\n");
	   ShowHelp ();
	   return 1;
	}

	*argv++;
	argc--;

	while (argc)
	{
	   if (!strcmp (*argv, "-a") || !strcmp (*argv, "--appointment-file"))
	   {
	      argc--;
	      *argv++;

	      if (!argc)
	      {
		 lfprintf (10001, stderr, "Parameter <-a> needs an argument!\n");
		 return 1;
	      }

	      strncpy (Kalendar, *argv, sizeof (Kalendar));
	      Kalendar[sizeof(Kalendar)-1] = 0;

	      if (strlen (Kalendar) < 1)
	      {
	         lfprintf (10002, stderr, "No valid file name!\n");
		 ShowHelp ();
	         return 1;
	      }
	   }

	   if (!strcmp (*argv, "-t") || !strcmp (*argv, "--local-time"))
	   {
	      argc--;
	      *argv++;
	      LocalTime = 1;
	      p = *argv;

	      if (argc && (*p == '-' || *p == '+') && isdigit (*(p+1)))
		 ZeitDifferenz = atoi (*argv);
	      else
	      {
		 argc++;
		 *argv--;
	      }
	   }

	   if (!strcmp (*argv, "-o") || !strcmp (*argv, "--output-file"))
	   {
	      argc--;
	      *argv++;

	      if (!argc)
	      {
	         lfprintf (10003, stderr, "The parameter <-o> needs an argument!\n");
		 ShowHelp ();
		 return 1;
	      }

	      strncpy (Outfile, *argv, sizeof (Outfile));
	      Outfile[sizeof (Outfile)-1] = 0;
	      Organizer = 1;
	   }

	   if (!strcmp (*argv, "-h") || !strcmp (*argv, "--help"))
	   {
	      ShowHelp ();
	      return 0;
	   }

	   if (!strcmp (*argv, "-w") || !strcmp (*argv, "--write-cell-phone"))
	      Handy = 1;

	   if (!strcmp (*argv, "-d") || !strcmp (*argv, "--device"))
	   {
	      argc--;
	      *argv++;

	      if (!argc)
	      {
		 lfprintf (10004, stderr, "The parameter <-d> needs an argument!\n");
		 ShowHelp ();
		 return 1;
	      }

	      strncpy (SerialDevice, *argv, sizeof (SerialDevice));
	      SerialDevice[sizeof(SerialDevice)-1] = 0;
	      Handy = 1;
	   }

	   if (!strcmp (*argv, "-b") || !strcmp (*argv, "--baud"))
	   {
	      argc--;
	      *argv++;

	      if (!argc)
	      {
	         lfprintf (10005, stderr, "The parameter <-b> needs an argument!\n");
		 ShowHelp ();
		 return 1;
	      }

	      /* Ueberpruefen der angegebenen Baudrate.
	       */
	      i = atoi (*argv);

	      switch (i)
	      {
	         case 19200:
		 case 38400:
		 case 57600:
		 case 115200:
		    Baudrate = i;
		 break;

		 default:
		    lfprintf (10006, stderr, "An illegal boud rate %s was given!\n", *argv);
	      }
	   }

	   if (!strcmp (*argv, "-p") || !strcmp (*argv, "--pin"))
	   {
	      argc--;
	      *argv++;

	      if (!argc)
	      {
	         lfprintf (10007, stderr, "The parameterr <-p> needs an argument!\n");
		 ShowHelp ();
		 return 1;
	      }

	      SetPin (*argv);
	   }

	   if (!strcmp (*argv, "-v") || !strcmp (*argv, "--verbose"))
	      Verbose = 1;

	   argc--;
	   *argv++;
	}

	if (Handy && !SerialDevice[0])
	   strcpy (SerialDevice, "/dev/ttyS0");

	if (!Kalendar[0])
	{
	   lfprintf (10008, stderr, "No source file found!\n");
	   ShowHelp ();
	   return 1;
	}

	if (Handy && !Baudrate)
	   Baudrate = 19200;

	/*
	 * Ueberpruefen ob ein Handy angeschlossen ist und ob es das
	 * erwartete Siemens S45 ist.
	 */
	if (Handy)
	{
	   if ((fdh = OpenHandy (SerialDevice, Baudrate)) == -1)
	      return 1;

	   if ((i = InitHandy (fdh, &rfrom, &rto)) < 0)
	   {
	      switch (i)
	      {
	         case -1:
		    lfprintf (10009, stderr, "An error occured!\n");
	         break;

		 case -2:
		    lfprintf (10010, stderr, "Unexpected answer!\n");
		 break;

		 case -4:
		    lfprintf (10011, stderr, "The connected device doesn't support a calendar!\n");
		 break;

		 default:
		    lfprintf (10012, stderr, "Unknown error %d.\n", i);
	      }

	      return 1;
	   }

	   if (rfrom < 0 || rto < 0 || rto <= rfrom)
	   {
	      lfprintf (10013, stderr, "The connected device doesn't have memory to store appointments!\n");
	      return 1;
	   }

	   GetManufacturer (fdh, &hv0[0]);
	   GetModel (fdh, &hv1[0]);
	   lprintf (10014, "Phone type : %s %s\n", hv0, hv1);

	   if (strcmp (hv0, "SIEMENS") || (strcmp (hv1, "S45") && strcmp (hv1, "S35")))
	   {
	      lprintf (10015, "This program was tested with the cellphones\n");
	      lprintf (10016, "\"SIEMENS S35/S45\" only! You've connected a device of type\n");
	      lprintf (10017, "%s %s, that may not work!\n\n", hv0, hv1);
	   }

	   if (Handy)
	   {
	      i = GetPhoneStatus (fdh);

	      if (i < 0)
	      {
	         lfprintf (10018, stderr, "Could not get phone status!\n");
		 lfprintf (10019, stderr, "Phone was deactivated!\n");
	         Handy = 0;
	      }
	      else
	      {
	         switch (i)
	         {
	            case 0: strcpy (hv0, "Ready"); break;
		    case 1: strcpy (hv0, "Incoming Call (Phone is ringing)"); break;
		    case 2: strcpy (hv0, "Call is active"); break;
		    default: strcpy (hv0, "Illegal or Unknown Status!");
	         }

	         lprintf (10020, "Phone status: %s\n", hv0);

	         if (i != 0)
	            Handy = 0;
	      }

	      //
	      // First delete all appointments to get full capacity of phone.
	      //
	      if (Handy)
	      {
	         zeile = rfrom;

	         while (zeile <= rto)
	         {
	            lprintf (10021, "Deleting slot: %2d ... ", zeile);

	            if ((i = DeleteEntry (fdh, zeile)) < 0)
	               puts ("EMPTY\r");
		    else
		       puts ("OK   \r");

		    fflush (stdout);
		    zeile++;
	         }

		 putchar ('\n');
		 fflush (NULL);
	      }
	      else
	         CloseHandy (fdh);
	   }
	   else
	      CloseHandy (fdh);
	}

	/*
	 * Oeffnen und einlesen des Dateiinhalts. Es wird so viel
	 * Speicher reserviert, wie die Datei lang ist.
	 */
	if ((fd = open (Kalendar, O_RDONLY, 0400)) == -1)
	{
	   lfprintf (10022, stderr, "Error opening file <%s>!\n", Kalendar);

	   if (Handy)
	      CloseHandy (fdh);

	   return 1;
	}

	if (fstat (fd, &st) == -1)
	{
	   lfprintf (10023, stderr, "Error stating file <%s>!\n", Kalendar);
	   close (fd);

	   if (Handy)
	      CloseHandy (fdh);

	   return 1;
	}

	if ((buffer = (char *)malloc (st.st_size+5)) == NULL)
	{
	   lfprintf (10024, stderr, "Not enough memory!\n");
	   close (fd);

	   if (Handy)
	      CloseHandy (fdh);

	   return 1;
	}

	memset (&vcal, 0, sizeof (vcal));
	memset (&vcal_head, 0, sizeof (vcal_head));
	vcal.x_pilotstat = 1;		// allways 1!

	read (fd, buffer, st.st_size);
	AvoidNull (buffer, st.st_size);
	buffer[st.st_size] = 0;
	pos = buffer;
	i = 1;

	/*
	 * Erzeugen der Zieldatei.
	 */
	if (Organizer)
	{
	   if ((fp = fopen (Outfile, "w+")) == NULL)
	   {
	      lfprintf (10025, stderr, "Error creating file <%s>!\n", Outfile);
	      close (fd);
	      free (buffer);

	      if (Handy)
	         CloseHandy (fdh);

	      return 1;
	   }
	}
	else
	   fp = NULL;

	/*
	 * Im Folgenden wird die Datei Zeilenweise gelesen und in ihre
	 * Komponenten zerlegt. Dabei wird das Schluesselwort, welches
	 * jeweils am Anfang der Zeile steht und mit einem Doppelpunkt
	 * terminiert ist, in ein Token umgewandelt. Danach wird die
	 * Funktion aufgerufen, welche fuer die Bearbeitung des Tokens
	 * zustaendig ist. Dieser Funktion wird der Zeilenrest und der
	 * Zeiger auf den Rest der Datei uebergeben.
	 */
	zeile = 0;

	while ((p = GetLine (pos)) != NULL)
	{

	   if (!*pos)
	   {
	      if ((pos = strchr (p, '\f')) != NULL)
	      {
	         *pos = 0;

		 vcal.description = WinToUnix (p);
		 pos++;

		 while (*pos && (*pos == '\r' || *pos == '\n'))
		    pos++;
	      }
	      else
	         pos = p;

	      /*
	       * Hier wurde das Ende eines Eintrags erreicht. Bevor wir
	       * den naechsten Eintrag bearbeiten, muessen wir die
	       * in der Struktur gespeicherten Daten in die Datei
	       * schreiben. Danach geben wir den Speicher frei und
	       * setzen alle Pointer und Variablen auf 0.
	       */
	      /* --- Eintrag schreiben --- */
	      if (Handy)
	         WriteHandy (fdh);

	      if (Organizer)
	         WriteVEvent (fp);

	      /* --- Speicher frei geben --- */
	      FreeVEvent ();

	      if (Organizer && !Handy)
	         lprintf (10026, "Wrote entry %d. Evaluated entry %d\r", wVcalNum, i);

	      if (Organizer && Handy)
	      {
	         lprintf (10027, "Wrote organizer entry %d and phone slot %d.\r", wVcalNum, SlotNum);
		 fflush (stdout);
	      }

	      if (!Organizer && Handy)
	      {
	         lprintf (10028, "Wrote phone slot %d and evaluated entry %d.\r", SlotNum, i);
		 fflush (stdout);
	      }

	      i++;
//	      pos = p;
	      continue;
	   }

	   if ((dot = strchr (pos, ':')) != NULL)
	   {
	   int j;

	      *dot = 0;
	      dot++;
	      j = 0;

	      while (Keys[j][0] != '\0')
	      {
	         if (strcasecmp (pos, Keys[j]) == 0)
		 {
		 int tok;

		    tok = j + TOKEN_BASE;
	   	    ltrim (dot);

		    switch (tok)
		    {
		       case POSTEDDATE:
			   WorkPostedDate (dot, p);
		       break;

		       case APPTUNID:
			   WorkAPPTUNID (dot, p);
		       break;

		       case SEQUENCENUM:
			   WorkSequenceNum (dot, p);
		       break;

		       case PRINCIPAL:
			   WorkINetSendTo (dot, p);
		       break;

		       case REQUIREDATTENDEES:
			   WorkINetRequiredNames (dot, p);
		       break;

		       case TOPIC:
		       case SUBJECT:
			   WorkSubject (dot, p);
		       break;

		       case PUBLICACCESS:
			   WorkPublicAccess (dot, p);
		       break;

		       case BUSYPRIORITY:
			   WorkBusyPriority (dot, p);
		       break;

		       case CATEGORIES:
			   WorkCategories (dot, p);
		       break;

		       case STARTTIME:
			   WorkStartTime (dot, p);
		       break;

		       case STARTDATETIME:
		           WorkStartDateTime (dot, p);
		       break;

		       case ENDDATETIME:
			   WorkEndDateTime (dot, p);
		       break;

		       case TYPE:
		           WorkType (dot, p);
		       break;

		       case ENDTIME:
			   WorkEndTime (dot, p);
		       break;

		       case ALARMS:
		           WorkAlarm (dot, p);
		       break;

		       case ALARMOFFSET:
		           WorkAlarmOffset (dot, p);
		       break;

		       case APPOINTMENTTYPE:
		           WorkAppointmentType (dot, p);
		       break;

		       case REPEATS:
		           WorkRepeats (dot, p);
		       break;

		       case LOCATION:
		           WorkLocation (dot, p);
		       break;

		       case STARTDATE:
		           WorkStartDate (dot, p);
		       break;

		       case ENDDATE:
		           WorkEndDate (dot, p);
		       break;

		       case STARTTIMEZONE:
		          WorkStartTimeZone (dot, p);
		       break;

		       case ENDTIMEZONE:
		          WorkEndTimeZone (dot, p);
		       break;
		    }
		 }

		 j++;
	      }
	   }
	   else if (*pos == '\x0c')
	   {
	      /*
	       * Hier wurde das Ende eines Eintrags erreicht. Bevor wir
	       * den naechsten Eintrag bearbeiten, muessen wir die
	       * in der Struktur gespeicherten Daten in die Datei
	       * schreiben. Danach geben wir den Speicher frei und
	       * setzen alle Pointer und Variablen auf 0.
	       */
	      /* --- Eintrag schreiben --- */
	      if (Handy)
	         WriteHandy (fdh);

	      if (Organizer)
	         WriteVEvent (fp);

	      /* --- Speicher frei geben --- */
	      FreeVEvent ();

	      if (Organizer && !Handy)
	         lprintf (10026, "Wrote entry %d.\r", i);

	      if (Organizer && Handy)
	      {
	         lprintf (10027, "Wrote organizer entry %d and phone slot %d.\r", i, SlotNum);
		 fflush (stdout);
	      }

	      if (!Organizer && Handy)
	      {
	         lprintf (10028, "Wrote phone slot %d.\r", SlotNum);
		 fflush (stdout);
	      }

	      i++;
	   }

	   pos = p;
	}

	close (fd);

	if (Organizer)
	{
	   fprintf (fp, "END:VCALENDAR\n");
	   fclose (fp);
	}

	free (buffer);
	putchar ('\n');
	fflush (NULL);

	if (Handy)
	   CloseHandy (fdh);

	FreeSlot ();
	return 0;
}

char *WorkComment (char *line, char *rest)
{
char *pos, *p, *buf, *jump;
long len;

	if ((pos = MyStrdup (rest)) == NULL)
	   return NULL;

	if ((p = strchr (pos, '\x0c')) != NULL)
	   *p = 0;

	len = (long)strlen (pos);
	jump = rest + len + 1;

	if ((buf = (char *)malloc (len * 2L)) == NULL)
	{
	   lfprintf (10029, stderr, "Not enough memory for <DESCRIPTION>!\n");
	   return pos;
	}

	if (vcal.description)
	   free (vcal.description);

	vcal.description = buf;
	*buf = 0;
	p = pos;

	while (p && *p)
	{
	   if (*p == '\r')
	   {
	      p++;
	      continue;
	   }

	   if (*p == '\n')
	   {
	      *buf = *p;
	      *(buf+1) = ' ';
	      buf += 2;
	      p++;
	      continue;
	   }

	   *buf = *p;
	   buf++;
	   p++;
	}

	free (pos);
	pos = WinToUnix (vcal.description);
	free (vcal.description);
	vcal.description = pos;
	return pos;
}

void WorkPostedDate (char *line, char *rest)
{
	vcal.created = ConvertDateTime (NULL, line);
	vcal.last_modified = ConvertDateTime (NULL, line);
	vcal.dtstamp = ConvertDateTime (NULL, line);
}

void WorkAPPTUNID (char *line, char *rest)
{
	vcal.UID = strdup (line);
}

void WorkSequenceNum (char *line, char *rest)
{
	vcal.sequence = atoi (line);
}

void WorkINetSendTo (char *line, char *rest)
{
	if (!line || !*line)
	   return;

	if (!strncasecmp (line, "CN=", 3))
	{
	struct _X400 *x400;

	   if ((x400 = (struct _X400 *)GetX400Address (line)) == NULL)
	      return;

	   vcal.organizer = x400->inet;

	   if (x400->cn)
	      free (x400->cn);

	   if (x400->ou)
	      free (x400->ou);

	   if (x400->o)
	      free (x400->o);

	   if (x400->c)
	      free (x400->c);

	   free (x400);
	}
	else if (strchr (line, '/') != NULL)
	   vcal.organizer = MyStrdup (GetNotesAddress (line));
	else
	   vcal.organizer = MyStrdup (line);

}

void WorkINetRequiredNames (char *line, char *rest)
{
char *p, *np, *xp, *ln;
struct _vcalendar_attendee *first, *akt;

	if (!line || *line == 0)
	   return;

	first = akt = NULL;

	if ((ln = MyStrdup (line)) == NULL)
	   return;

	p = ln;

	if ((np = strchr (p, ',')) != NULL)
	   *np = 0;
	else
	   np = p;

	while (p)
	{
	   if (!first)
	   {
	      if ((first = (struct _vcalendar_attendee *)malloc (sizeof (struct _vcalendar_attendee))) == NULL)
	      {
		 lfprintf (10030, stderr, "Not enough memory for <ATTENDEE>!\n");
		 free (ln);
		 return;
	      }

	      akt = first;
	   }
	   else
	   {
	   struct _vcalendar_attendee *tp;

	      if ((tp = (struct _vcalendar_attendee *)malloc (sizeof (struct _vcalendar_attendee))) == NULL)
	      {
		 lfprintf (10030, stderr, "Not enough memory for <ATTENDEE>!\n");
		 vcal.vcal_att = first;
		 free (ln);
		 return;
	      }

	      akt->next = tp;
	      akt = tp;
	   }

	   memset (akt, 0, sizeof (struct _vcalendar_attendee));

	   /*
	    * An dieser Stelle zerlegen wir einen "LDAP"-String. Notes
	    * speichert seine Mailadressen machmal in LDAP-Notation.
	    * Manchmal auch in einer durch slash getrennte Teile und
	    * machmal auch in E-Mail-Notation.
	    * Es muss also geprueft werden, welche Notation von Notes
	    * verwendet wurde!
	    */
	   if (!strncasecmp (p, "CN=", 3))	// X400 Notation?
	   {
	   struct _X400 *x400;

	      if ((x400 = GetX400Address (p)) == NULL)
	      {
	         free (ln);
		 lfprintf (10082, stderr, "Error parsing a X400 address >%s<!\n", p);
		 return;
	      }

	      akt->cn = x400->cn;
	      akt->mailto = x400->inet;

	      if (x400->ou)
	         free (x400->ou);

	      if (x400->o)
	         free (x400->o);

	      if (x400->c)
	         free (x400->c);

	      free (x400);
	   }
	   else if ((xp = strchr (p, '/')) != NULL)
	   {
	      *xp = 0;

	      if ((akt->cn = MyStrdup (p)) == NULL)
	      {
	         free (ln);
		 return;
	      }

	      *xp = '/';
	      xp = GetNotesAddress (p);

	      if ((akt->mailto = MyStrdup (xp)) == NULL)
	      {
	         free (ln);
		 return;
	      }
	   }
	   else
	   {
	      if ((xp = strchr (p, '@')) != NULL)
	      {
	         *xp = 0;

		 if ((akt->cn = MyStrdup (p)) == NULL)
		 {
		    free (ln);
		    return;
		 }

		 *xp = '@';
	      }

	      if ((akt->mailto = MyStrdup (p)) == NULL)
	      {
	         free (ln);
		 return;
	      }
	   }

	   akt->rsvpt = 0;
	   strcpy (akt->partstat, "ACCEPTED");
	   strcpy (akt->role, "REQ-PARTICIPANT");

	   if (p == np)
	      break;

	   np++;
	   p = np;

	   if ((np = strchr (p, ',')) != NULL)
	      *np = 0;
	   else
	      np = p;
	}


	free (ln);
	vcal.vcal_att = first;
}

void WorkSubject (char *line, char *rest)
{
	if (!vcal.summary)
	{
	   vcal.summary = WinToUnix (line);
#ifdef DEBUG
	   fprintf (stderr, "Subject: %s\n", vcal.summary);
#endif
	}
}

void WorkPublicAccess (char *line, char *rest)
{
int acc;

	acc = atoi (line);

	switch (acc)
	{
	   case 1: strcpy (vcal.class, "PUBLIC"); break;
	   default: strcpy (vcal.class, "PRIVATE"); break;
	}
}

void WorkBusyPriority (char *line, char *rest)
{
int prio;

	prio = atoi (line);
	vcal.priority = prio + 2;
}

void WorkCategories (char *line, char *rest)
{
	if (vcal.categories)
	   return;

	if (!*line || strlen (line) == 0)
	   vcal.categories = strdup ("Besprechung");
	else
	   vcal.categories = WinToUnix (line);
}

void WorkStartTime (char *line, char *rest)
{
struct _vcalendar_datetime *dt;

#ifdef DEBUG
	fprintf (stderr, "WorkStartTime: %s --> ", line);
	fflush (NULL);
#endif
	if ((dt = ConvertDateTime (NULL, line)) == NULL)
	   return;

	if (vcal.dtstart)
	{
	   if (dt->year > 0)
	   {
	      vcal.dtstart->year = dt->year;
	      vcal.dtstart->mon = dt->mon;
	      vcal.dtstart->day = dt->day;
	   }

	   vcal.dtstart->hour = dt->hour;
	   vcal.dtstart->min = dt->min;
	   vcal.dtstart->sec = dt->sec;
	   free (dt);
	}
	else
	   vcal.dtstart = dt;

#ifdef DEBUG
	fprintf (stderr, "%02d:%02d:%02d\n",
		vcal.dtstart->hour, vcal.dtstart->min, vcal.dtstart->sec);
#endif
}

void WorkStartDateTime (char *line, char *rest)
{
char *p, *ln;
int anz;
struct _vcalendar_datetime *dt;
struct _vcalendar_chaindt *akt, *prev;

	if (!line || strlen (line) < 19)
	{
	   lfprintf (10080, stderr, "Illegal date detected!\n");
	   return;
	}

#ifdef DEBUG
	fprintf (stderr, "WorkStartDateTime: %s\n", line);
	fflush (NULL);
#endif

	ln = strdup (line);

	if (!strchr (ln, ','))
	{
	   vcal.dstart = ConvertDateTime (NULL, ln);
	   free (ln);
	   return;
	}

	p = strtok (ln, ",");
	akt = NULL;
	anz = 0;

	if (vcal.chstart)	// This should never be true, but ...
	{
	   akt = vcal.chstart;

	   while (akt)
	   {
	      prev = akt->next;
	      free (akt);
	      akt = prev;
	   }

	   vcal.chstart = NULL;
	}

	while (p)
	{
	   if (!p || strlen (p) != 19)
	   {
#ifdef DEBUG
	      fprintf (stderr, "Ungueltiges Datum entdeckt: %s\n", p);
	      fflush (NULL);
#endif
	      lfprintf (10080, stderr, "Illegal date detected!\n");
	      free (ln);
	      return;
	   }

	   if ((dt = ConvertDateTime (NULL, p)) == NULL)
	   {
	      p = strtok (NULL, ",");
	      continue;
	   }

	   if (vcal.chstart == NULL)
	   {
	      if ((akt = (struct _vcalendar_chaindt *)malloc (sizeof (struct _vcalendar_chaindt))) == NULL)
	      {
	         lfprintf (10024, stderr, "Not enough memory!\n");
		 free (ln);
		 return;
	      }

	      akt->next = NULL;
	      vcal.chstart = akt;
	   }
	   else
	   {
	      prev = akt;

	      if ((akt = (struct _vcalendar_chaindt *)malloc (sizeof (struct _vcalendar_chaindt))) == NULL)
	      {
	         lfprintf (10024, stderr, "Not enough memory!\n");
		 free (ln);
		 return;
	      }

	      akt->next = NULL;
	      prev->next = akt;
	   }

	   akt->day = dt->day;
	   akt->mon = dt->mon;
	   akt->year = dt->year;
	   akt->hour = dt->hour;
	   akt->min = dt->min;
	   akt->sec = dt->sec;
	   anz++;
	   free (dt);

	   p = strtok (NULL, ",");
	}

	free (ln);
	vcal.chstart_anz = anz;
}

void WorkEndTime (char *line, char *rest)
{
struct _vcalendar_datetime *dt;

#ifdef DEBUG
	fprintf (stderr, "WorkEndTime: %s --> ", line);
	fflush (NULL);
#endif
	if ((dt = ConvertDateTime (NULL, line)) == NULL)
	   return;

	if (vcal.dtend)
	{
	   if (dt->year > 0)
	   {
	      vcal.dtend->year = dt->year;
	      vcal.dtend->mon = dt->mon;
	      vcal.dtend->day = dt->day;
	   }

	   vcal.dtend->hour = dt->hour;
	   vcal.dtend->min = dt->min;
	   vcal.dtend->sec = dt->sec;
	   free (dt);
	}
	else
	   vcal.dtend = dt;

#ifdef DEBUG
	fprintf (stderr, "%02d:%02d:%02d\n",
		vcal.dtend->hour, vcal.dtend->min, vcal.dtend->sec);
#endif
}

void WorkEndDateTime (char *line, char *rest)
{
char *p, *ln;
int anz;
struct _vcalendar_datetime *dt;
struct _vcalendar_chaindt *akt, *prev;

	if (!line || strlen (line) < 19)
	{
	   lfprintf (10080, stderr, "Illegal date detected!\n");
	   return;
	}

#ifdef DEBUG
	fprintf (stderr, "WorkEndDateTime: %s\n", line);
	fflush (NULL);
#endif
	ln = strdup (line);

	if (!strchr (ln, ','))
	{
	   vcal.dend = ConvertDateTime (NULL, ln);
	   free (ln);
	   return;
	}

	p = strtok (ln, ",");
	akt = NULL;
	anz = 0;

	if (vcal.chend)		// This should never be true, but ...
	{
	   akt = vcal.chend;

	   while (akt)
	   {
	      prev = akt->next;
	      free (akt);
	      akt = prev;
	   }

	   vcal.chend = NULL;
	}

	while (p)
	{
	   if (!p || strlen (p) != 19)
	   {
#ifdef DEBUG
	      fprintf (stderr, "Ungueltiges Datum entdeckt: %s\n", p);
	      fflush (NULL);
#endif
	      lfprintf (10080, stderr, "Illegal date detected!\n");
	      free (ln);
	      return;
	   }

	   if ((dt = ConvertDateTime (NULL, p)) == NULL)
	   {
	      p = strtok (NULL, ",");
	      continue;
	   }

	   if (vcal.chend == NULL)
	   {
	      if ((akt = (struct _vcalendar_chaindt *)malloc (sizeof (struct _vcalendar_chaindt))) == NULL)
	      {
	         lfprintf (10024, stderr, "Not enough memory!\n");
		 free (ln);
		 return;
	      }

	      akt->next = NULL;
	      vcal.chend = akt;
	   }
	   else
	   {
	      prev = akt;

	      if ((akt = (struct _vcalendar_chaindt *)malloc (sizeof (struct _vcalendar_chaindt))) == NULL)
	      {
	         lfprintf (10024, stderr, "Not enough memory!\n");
		 free (ln);
		 return;
	      }

	      akt->next = NULL;
	      prev->next = akt;
	   }

	   akt->day = dt->day;
	   akt->mon = dt->mon;
	   akt->year = dt->year;
	   akt->hour = dt->hour;
	   akt->min = dt->min;
	   akt->sec = dt->sec;
	   anz++;
	   free (dt);

	   p = strtok (NULL, ",");
	}

	free (ln);
	vcal.chend_anz = anz;
}

void WorkType (char *line, char *rest)
{
	vcal.type = strdup (line);

	if (!vcal.categories)
	   vcal.categories = WinToUnix (line);
}

void WorkAlarm (char *line, char *rest)
{
	if (*line)
	   vcal.alarm = atoi (line);
	else
	   vcal.alarm = 0;
}

void WorkRepeats (char *line, char *rest)
{
	if (*line)
	   vcal.repeats = atoi (line);
	else
	   vcal.repeats = 0;
}

void WorkAlarmOffset (char *line, char *rest)
{
	if (*line)
	   vcal.alarm_offset = atoi (line);
}

void WorkAppointmentType (char *line, char *rest)
{
int at;

	if (!*line)
	   return;

	at = atoi (line);
	vcal.appointment_type = at;

	if (vcal.categories)
	   free (vcal.categories);

	switch (at)
	{
	   case 0:
	      vcal.categories = strdup ("Termin");
	   break;

	   case 1:
	      vcal.categories = strdup ("Feiertag");
	   break;

	   case 2:
	      vcal.categories = strdup ("Ganzt\xc3\xa4gige Veranstaltung");
	   break;

	   case 3:
	      vcal.categories = strdup ("Besprechung");
	   break;

	   default:
	      vcal.categories = strdup ("Besprechung");
	}
}

void WorkLocation (char *line, char *rest)
{
	vcal.location = WinToUnix (line);
}

void WorkStartDate (char *line, char *rest)
{
struct _vcalendar_datetime *dt;

	if (!(dt = ConvertDateTime (NULL, line)))
	   return;

	if (vcal.dtstart)
	{
	   vcal.dtstart->year = dt->year;
	   vcal.dtstart->mon = dt->mon;
	   vcal.dtstart->day = dt->day;

	   if (dt->hour != 0 || dt->min != 0 || dt->sec != 0)
	   {
	      vcal.dtstart->hour = dt->hour;
	      vcal.dtstart->min = dt->min;
	      vcal.dtstart->sec = dt->sec;
	   }

	   free (dt);
	}
	else
	   vcal.dtstart = dt;
}

void WorkEndDate (char *line, char *rest)
{
struct _vcalendar_datetime *dt;

	if ((dt = ConvertDateTime (NULL, line)) == NULL)
	   return;

	if (vcal.dtend)
	{
	   vcal.dtend->year = dt->year;
	   vcal.dtend->mon = dt->mon;
	   vcal.dtend->day = dt->day;

	   if (dt->hour != 0 || dt->min != 0 || dt->sec != 0)
	   {
	      vcal.dtend->hour = dt->hour;
	      vcal.dtend->min = dt->min;
	      vcal.dtend->sec = dt->sec;
	   }

	   free (dt);
	}
	else
	   vcal.dtend = dt;
}

void WorkStartTimeZone (char *line, char *rest)
{
	vcal.tzstart = GetTimeZone (line);
}

void WorkEndTimeZone (char *line, char *rest)
{
	vcal.tzend = GetTimeZone (line);
}

void WriteVEvent (FILE *fp)
{
time_t t;
int year;
static int flag = 0;
struct tm *lt;

	if (!vcal.summary)
	{
	   if (Verbose)
	   {
	      if (vcal.UID)
	         lprintf (10084, "Found entry without \"Subject\"! (UID: %s)\n", vcal.UID);
	      else
	         lprintf (10075, "Found entry without \"Subject\"!\n");
	   }

	   if ((vcal.summary = MyStrdup ("Missing subject")) == NULL)
	      return;
	}

	if (!vcal.UID && flag)
	   return;

	if (!vcal.UID && !flag)
	{
	   if (Verbose)
	      lprintf (10076, "Found entry (%s) without APPTUNID!\n", vcal.summary);

	   flag = 1;
	}

	if (vcal.UID)
	   flag = 0;

	if (!vcal_head.prodid[0])
	{
	   fprintf (fp, "BEGIN:VCALENDAR\n");
	   strcpy (vcal_head.prodid, "-//K Desktop Environment//NONSGML KOrganizer 3.0.3//EN");
	   strcpy (vcal_head.version, "2.0");
	   fprintf (fp, "PRODID\n :%s\nVERSION\n :%s\n", vcal_head.prodid, vcal_head.version);
	}

	t = time (NULL);
	lt = localtime (&t);

	if (lt->tm_year < 1900)
	   year = lt->tm_year + 1900;
	else
	   year = lt->tm_year;

	/*
	 * Eintraege die aelter als das aktuelle Jahr sind,
	 * werden ignoriert, sofern es sich nicht um einen
	 * sich widerholenden Eintrag handelt.
	 */
	if (vcal.dtstart && vcal.dtstart->year < year && vcal.chstart_anz == 0)
	   return;

	/*
	 * Wenn der aktuelle Eintrag ein Feiertag ist, dann hat er ein
	 * sich wiederholendes Datum. Hier rufen wir nun diese Funktion
	 * fr jedes Datum rekursiv auf. Damit die Sache zu keiner
	 * Endlosschleife wird, setzen wir den Zeiger auf die Daten
	 * auf NULL.
	 */
	if (vcal.chstart_anz > 1 && vcal.chstart)
	{
	struct _vcalendar_chaindt *dt, *sakt, *eakt;
	char hv0[512];
	int i;

	   dt = sakt = vcal.chstart;
	   eakt = vcal.chend;
	   vcal.chstart = NULL;	// Notwendig um keine Endlosschleife zu machen!
	   i = 0;

	   while (sakt)
	   {
	      if (!vcal.dtstart)
	      {
	         if ((vcal.dtstart = (struct _vcalendar_datetime *)malloc (sizeof (struct _vcalendar_datetime))) == NULL)
		 {
		    vcal.dtstart = NULL;
		    lfprintf (10024, stderr, "Not enough memory!\n");
		    break;
		 }
	      }

	      if (!vcal.dtend)
	      {
	         if ((vcal.dtend = (struct _vcalendar_datetime *)malloc (sizeof (struct _vcalendar_datetime))) == NULL)
		 {
		    vcal.dtend = NULL;
		    lfprintf (10024, stderr, "Not enough memory!\n");
		    break;
		 }
	      }

	      if (vcal.UID && strlen (vcal.UID) < (sizeof (hv0) - 10))
	      {
	      char *p;

	         if ((p = strchr (vcal.UID, '.')) != NULL)
		    *p = 0;

	         sprintf (hv0, "%s.%03d", vcal.UID, i);
	         free (vcal.UID);
	         vcal.UID = strdup (hv0);
	      }

	      if (eakt)
	      {
	         vcal.dtend->day = eakt->day;
	         vcal.dtend->mon = eakt->mon;
	         vcal.dtend->year = eakt->year;
	         vcal.dtend->hour = eakt->hour;
	         vcal.dtend->min = eakt->min;
	         vcal.dtend->sec = eakt->sec;
	      }
	      else
	      {
	         vcal.dtend->day = sakt->day;
	         vcal.dtend->mon = sakt->mon;
	         vcal.dtend->year = sakt->year;
	         vcal.dtend->hour = sakt->hour;
	         vcal.dtend->min = sakt->min;
	         vcal.dtend->sec = sakt->sec;
	      }

	      vcal.dtstart->day = sakt->day;
	      vcal.dtstart->mon = sakt->mon;
	      vcal.dtstart->year = sakt->year;
	      vcal.dtstart->hour = sakt->hour;
	      vcal.dtstart->min = sakt->min;
	      vcal.dtstart->sec = sakt->sec;
	      WriteVEvent (fp);
	      sakt = sakt->next;
	      i++;

	      if (eakt)
	         eakt = eakt->next;
	   }

	   vcal.chstart = dt;
	   return;
	}

	if (vcal.chstart_anz == 1 && !vcal.dtstart && vcal.chstart)
	{
	   if ((vcal.dtstart = (struct _vcalendar_datetime *)malloc (sizeof (struct _vcalendar_datetime))) == NULL)
	   {
	      lfprintf (10024, stderr, "Not enough memory!\n");
	      return;
	   }

	   vcal.dtstart->day = vcal.chstart->day;
	   vcal.dtstart->mon = vcal.chstart->mon;
	   vcal.dtstart->year = vcal.chstart->year;
	   vcal.dtstart->hour = vcal.chstart->hour;
	   vcal.dtstart->min = vcal.chstart->min;
	   vcal.dtstart->sec = vcal.chstart->sec;
	}

	if (vcal.chend_anz == 1 && !vcal.dtend && vcal.chend)
	{
	   if ((vcal.dtend = (struct _vcalendar_datetime *)malloc (sizeof (struct _vcalendar_datetime))) == NULL)
	   {
	      lfprintf (10024, stderr, "Not enough memory!\n");
	      return;
	   }

	   vcal.dtend->day = vcal.chend->day;
	   vcal.dtend->mon = vcal.chend->mon;
	   vcal.dtend->year = vcal.chend->year;
	   vcal.dtend->hour = vcal.chend->hour;
	   vcal.dtend->min = vcal.chend->min;
	   vcal.dtend->sec = vcal.chend->sec;
	}

	fprintf (fp, "BEGIN:VEVENT\n");

	if (vcal.created)
	   fprintf (fp, "CREATED\n :%04d%02d%02dT%02d%02d%02dZ\n",
	   	vcal.created->year, vcal.created->mon, vcal.created->day,
		vcal.created->hour, vcal.created->min, vcal.created->sec);
	else
	   fprintf (fp, "CREATED\n :%04d%02d%02dT%02d%02d%02dZ\n",
	   	(lt->tm_year < 1900) ? lt->tm_year + 1900 : lt->tm_year,
		lt->tm_mon + 1, lt->tm_mday, lt->tm_hour, lt->tm_min, lt->tm_sec);

	if (vcal.UID)
	   fprintf (fp, "UID\n :%s\n", vcal.UID);

	fprintf (fp, "SEQUENCE\n :%d\n", vcal.sequence);

	if (vcal.last_modified)
	   fprintf (fp, "LAST-MODIFIED\n :%04d%02d%02dT%02d%02d%02dZ\n",
	   	vcal.last_modified->year, vcal.last_modified->mon, vcal.last_modified->day,
		vcal.last_modified->hour, vcal.last_modified->min, vcal.last_modified->sec);
	else
	   fprintf (fp, "LAST-MODIFIED\n :%04d%02d%02dT%02d%02d%02dZ\n",
	   	(lt->tm_year < 1900) ? lt->tm_year + 1900 : lt->tm_year,
		lt->tm_mon + 1, lt->tm_mday, lt->tm_hour, lt->tm_min, lt->tm_sec);

	if (vcal.dtstamp)
	   fprintf (fp, "DTSTAMP\n :%04d%02d%02dT%02d%02d%02dZ\n",
	   	vcal.dtstamp->year, vcal.dtstamp->mon, vcal.dtstamp->day,
		vcal.dtstamp->hour, vcal.dtstamp->min, vcal.dtstamp->sec);

	else
	   fprintf (fp, "DTSTAMP\n :%04d%02d%02dT%02d%02d%02dZ\n",
	   	(lt->tm_year < 1900) ? lt->tm_year + 1900 : lt->tm_year,
		lt->tm_mon + 1, lt->tm_mday, lt->tm_hour, lt->tm_min, lt->tm_sec);

	if (vcal.organizer)
	   fprintf (fp, "ORGANIZER\n :MAILTO:%s\n", vcal.organizer);

	if (vcal.vcal_att)
	{
	struct _vcalendar_attendee *akt;

	   akt = vcal.vcal_att;

	   while (akt)
	   {
	      fprintf (fp, "ATTENDEE\n ;CN=%s\n ;RSVP=%s\n ;PARTSTAT=%s\n ;ROLE=%s\n :mailto:%s\n",
		akt->cn, (akt->rsvpt == 0) ? "FALSE" : "TRUE", akt->partstat, akt->role, akt->mailto);
	      akt = akt->next;
	   }
	}

	if (vcal.description)
	   fprintf (fp, "DESCRIPTION\n :%s\n", vcal.description);

	if (vcal.location)
	   fprintf (fp, "LOCATION\n :%s\n", vcal.location);

	if (vcal.summary)
	   fprintf (fp, "SUMMARY\n :%s\n", vcal.summary);

	fprintf (fp, "CLASS\n :%s\nPRIORITY\n :%d\n", vcal.class, vcal.priority);

	if (vcal.categories)
	   fprintf (fp, "CATEGORIES\n :%s\n", vcal.categories);

	fprintf (fp, "X-PILOTID\n :%d\nX-PILOTSTAT\n :%d\n", vcal.x_pilotid, vcal.x_pilotstat);

	if (vcal.duration)
	   fprintf (fp, "DURATION\n :%s\n", vcal.duration);
	else
	   fprintf (fp, "DURATION\n :PTS0\n");

	if (vcal.dtstart)
	{
	   if (vcal.appointment_type == 1 || vcal.appointment_type == 2)
	   {
	      fprintf (fp, "DTSTART\n ;VALUE=DATE\n :%04d%02d%02d\n",
	      	vcal.dtstart->year, vcal.dtstart->mon, vcal.dtstart->day);
	   }
	   else
	   {
	      TimeZone (vcal.dtstart, false);
	      fprintf (fp, "DTSTART\n :%04d%02d%02dT%02d%02d%02dZ\n",
	   	vcal.dtstart->year, vcal.dtstart->mon, vcal.dtstart->day,
		vcal.dtstart->hour, vcal.dtstart->min, vcal.dtstart->sec);
	   }
	}
#ifdef DEBUG
	else
	   fprintf (stderr, "vcal.dtstart ist NULL >%s<\n", vcal.summary);
#endif

	if (!vcal.dtstart && vcal.dstart)
	{
	   fprintf (fp, "DTSTART\n ;VALUE=DATE\n :%04d%02d%02d\n",
	      	vcal.dstart->year, vcal.dstart->mon, vcal.dstart->day);
	}

	if (vcal.dtend)
	{
	   if (vcal.appointment_type == 1 || vcal.appointment_type == 2)
	   {
	      fprintf (fp, "DTEND\n ;VALUE=DATE\n :%04d%02d%02d\n",
	      	vcal.dtend->year, vcal.dtend->mon, vcal.dtend->day);
	   }
	   else
	   {
	      TimeZone (vcal.dtend, false);
	      fprintf (fp, "DTEND\n :%04d%02d%02dT%02d%02d%02dZ\n",
	   	vcal.dtend->year, vcal.dtend->mon, vcal.dtend->day,
		vcal.dtend->hour, vcal.dtend->min, vcal.dtend->sec);
	   }
	}

	if (!vcal.dtend && vcal.dend)
	{
	   fprintf (fp, "DTEND\n ;VALUE=DATE\n :%04d%02d%02d\n",
	      	vcal.dend->year, vcal.dend->mon, vcal.dend->day);
	}

	if (vcal.vcal_valarm)
	{
	   fprintf (fp, "BEGIN:VALARM\nACTION\n :%s\n", vcal.vcal_valarm->action);

	   if (vcal.vcal_valarm->vcal_trigger)
	   {
	      fprintf (fp, "TRIGGER\n");

	      if (vcal.vcal_valarm->vcal_trigger->value)
		 fprintf (fp, " ;%s\n", vcal.vcal_valarm->vcal_trigger->value);

	      if (vcal.vcal_valarm->vcal_trigger->duration)
		 fprintf (fp, " :%s\n", vcal.vcal_valarm->vcal_trigger->duration);
	   }

	   fprintf (fp, "END:VALARM");
	}

	fprintf (fp, "END:VEVENT\n");
	wVcalNum++;
}

void WriteHandy (int fd)
{
char hv0[512];
char *p;
char Data[4096];
time_t t, dtstart, dtend;
int i, dif, repeat;
static int flag = 0;
struct tm lt, *tp, *gmt;

	dtstart = dtend = 0;

	if (!vcal.summary || SlotNum > rto || vcal.appointment_type == 1)
	   return;

	if (!vcal.UID && flag)
	   return;

	if (!vcal.UID && !flag)
	{
	   if (Verbose)
	      lprintf (10076, "Found entry (%s) without APPTUNID!\n", vcal.summary);

	   if (vcal.appointment_type != 2)
	      flag = 1;
	}

	if (vcal.UID)
	   flag = 0;

	t = time (NULL);
	tp = localtime (&t);
/*	memcpy (&lt, tp, sizeof (struct tm));
	gmt = gmtime (&t);
	dif = 0;

	if (ZeitDifferenz)
	   dif = ZeitDifferenz;
	else if (!LocalTime)
	   dif = lt.tm_hour - gmt->tm_hour;
*/
	/*
	 * Eintraege die aelter als das aktuelle Datum sind,
	 * werden ignoriert. Ausgenommen es handelt sich um
	 * einen sich wiederholenden Termin.
	 */
	memset (&lt, 0, sizeof (struct tm));

	if (vcal.dtstart)
	{
	   lt.tm_year = vcal.dtstart->year - 1900;
	   lt.tm_mon = vcal.dtstart->mon - 1;
	   lt.tm_mday = vcal.dtstart->day;
//	   lt.tm_hour = vcal.dtstart->hour + dif;
	   lt.tm_hour = vcal.dtstart->hour;
	   lt.tm_min = vcal.dtstart->min;
	   lt.tm_sec = vcal.dtstart->sec;
	   dtstart = mktime (&lt);
	}
	else if (vcal.dstart)
	{
	   lt.tm_year = vcal.dstart->year - 1900;
	   lt.tm_mon = vcal.dstart->mon - 1;
	   lt.tm_mday = vcal.dstart->day;
//	   lt.tm_hour = vcal.dstart->hour + dif;
	   lt.tm_hour = vcal.dstart->hour;
	   lt.tm_min = vcal.dstart->min;
	   lt.tm_sec = vcal.dstart->sec;
	   dtstart = mktime (&lt);
	}
	else
	{
	   if (Verbose)
	      lprintf (10077, "Entry has no valid start date!\n");

	   return;
	}

	if (!vcal.dtend && !vcal.dend)
	{
	   if (Verbose)
	      lprintf (10078, "Entry has no valid end date!\n");

	   return;
	}

	if (dtstart < t && !vcal.repeats)
	   return;

	/*
	 * Um sicher zu gehen, das sich wiederholende Termine deren letzte
	 * Wiederholung schon in der Vergangenheit liegt nicht gesendet
	 * werden, pruefen wir das letzte Endedatum.
	 */
	if (vcal.repeats && vcal.chend && vcal.chend_anz > 1)
	{
	struct _vcalendar_chaindt *akt, *ldate;

	   akt = vcal.chend;

	   while (akt)
	   {
	      ldate = akt;
	      akt = akt->next;
	   }

	   lt.tm_year = ldate->year - 1900;
	   lt.tm_mon = ldate->mon - 1;
	   lt.tm_mday = ldate->day;
//	   lt.tm_hour = ldate->hour + dif;
	   lt.tm_hour = ldate->hour;
	   lt.tm_min = ldate->min;
	   lt.tm_sec = ldate->sec;
	   dtend = mktime (&lt);

	   if (dtend < t)
	      return;
	}

	/*
	 * Da Lotus Notes leider einige Eintraege mehrfach exportiert,
	 * muessen wir in unserem "Slotspeicher" nachsehen, ob der
	 * vorliegende Eintrag vielleicht schon existiert.
	 */
	if (!dtend)
	{
	   if (vcal.dtend)
	   {
	      lt.tm_year = vcal.dtend->year - 1900;
	      lt.tm_mon = vcal.dtend->mon - 1;
	      lt.tm_mday = vcal.dtend->day;
//	      lt.tm_hour = vcal.dtend->hour + dif;
	      lt.tm_hour = vcal.dtend->hour;
	      lt.tm_min = vcal.dtend->min;
	      lt.tm_sec = vcal.dtend->sec;
	      dtend = mktime (&lt);
	   }
	   else if (vcal.dend)
	   {
	      lt.tm_year = vcal.dend->year - 1900;
	      lt.tm_mon = vcal.dend->mon - 1;
	      lt.tm_mday = vcal.dend->day;
//	      lt.tm_hour = vcal.dend->hour + dif;
	      lt.tm_hour = vcal.dend->hour;
	      lt.tm_min = vcal.dend->min;
	      lt.tm_sec = vcal.dend->sec;
	      dtend = mktime (&lt);
	   }
	}

	if (FindSlot (dtstart, dtend, vcal.summary) != NULL)
	{
	   if (Verbose)
	      lprintf (10079, "Double entry \"%s\" was found and ignored.\n", vcal.summary);

	   return;
	}

	if (vcal.repeats && FindSlotSummary (vcal.summary) != NULL)
	{
	   if (Verbose)
	      lprintf (10085, "Potential double entry \"%s\" was found and ignored!\n", vcal.summary);

	   return;
	}

	/*
	 * Das Startdatum ist also >= dem aktuellen Datum. Wir koennen
	 * den Eintrag zusammenstellen...
	 */
	strcpy (Data, "BEGIN:VCALENDAR\r\nVERSION:1.0\r\nBEGIN:VEVENT\r\n");
	strcat (Data, "CATEGORIES:");

	switch (vcal.appointment_type)
	{
	   case 0:
	      strcat (Data, "MEETING\r\n");
	   break;

	   case 1:
	      strcat (Data, "MISCELLANEOUS\r\n");
	   break;

	   case 2:
	      strcat (Data, "MEETING\r\n");
	   break;

	   case 3:
	      strcat (Data, "MEETING\r\n");
	   break;

	   default:
	      strcat (Data, "MEETING\r\n");
	}

	if (vcal.alarm == 1 && vcal.dtstart)
	{
	   lt.tm_year = vcal.dtstart->year - 1900;
	   lt.tm_mon = vcal.dtstart->mon - 1;
	   lt.tm_mday = vcal.dtstart->day;
//	   lt.tm_hour = vcal.dtstart->hour + dif;
	   lt.tm_hour = vcal.dtstart->hour;
	   lt.tm_min = vcal.dtstart->min;
	   lt.tm_sec = vcal.dtstart->sec;

	   if (vcal.alarm_offset < 0)
	   {
	      vcal.alarm_offset /= -1;
	      t = mktime (&lt) - ((time_t)vcal.alarm_offset * (time_t)60);
	   }
	   else
	      t = mktime (&lt) + ((time_t)vcal.alarm_offset * (time_t)60);

	   tp = localtime (&t);

	   sprintf (hv0, "DALARM:%04d%02d%02dT%02d%02d%02dZ\r\n",
	   	tp->tm_year+1900, tp->tm_mon+1, tp->tm_mday,
		tp->tm_hour, tp->tm_min, tp->tm_sec);

	   strcat (Data, hv0);
	}

	if (vcal.dtstart)
	{
	   if (vcal.appointment_type == 1 || vcal.appointment_type == 2)
	      sprintf (hv0, "DTSTART:%04d%02d%02dT000000Z\r\n",
	      	vcal.dtstart->year, vcal.dtstart->mon, vcal.dtstart->day);
	   else
	      sprintf (hv0, "DTSTART:%04d%02d%02dT%02d%02d%02dZ\r\n",
	   	vcal.dtstart->year, vcal.dtstart->mon, vcal.dtstart->day,
		vcal.dtstart->hour, vcal.dtstart->min, vcal.dtstart->sec);
//		vcal.dtstart->hour + dif, vcal.dtstart->min, vcal.dtstart->sec);

	   strcat (Data, hv0);
	}

	if (!vcal.dtstart && vcal.dstart)
	{
	   sprintf (hv0, "DTSTART:%04d%02d%02dT000000Z\r\n",
	      	vcal.dstart->year, vcal.dstart->mon, vcal.dstart->day);

	   strcat (Data, hv0);
	}

	if (vcal.dtend)
	{
	struct _vcalendar_chaindt *akt, *ldate;

	   if (vcal.chend && vcal.repeats)
	   {
	      akt = vcal.chend;

	      while (akt)
	      {
	         ldate = akt;
	         akt = akt->next;
	      }

	      lt.tm_year = ldate->year - 1900;
	      lt.tm_mon = ldate->mon - 1;
	      lt.tm_mday = ldate->day;
//	      lt.tm_hour = ldate->hour + dif;
	      lt.tm_hour = ldate->hour;
	      lt.tm_min = ldate->min;
	      lt.tm_sec = ldate->sec;
	      dtend = mktime (&lt);

	      if (vcal.appointment_type == 1 || vcal.appointment_type == 2)
	         sprintf (hv0, "DTEND:%04d%02d%02dT235959Z\r\n",
	      	   ldate->year, ldate->mon, ldate->day);
	      else
	         sprintf (hv0, "DTEND:%04d%02d%02dT%02d%02d%02dZ\r\n",
	   	   ldate->year, ldate->mon, ldate->day,
		   ldate->hour, ldate->min, ldate->sec);
//		   ldate->hour + dif, ldate->min, ldate->sec);
	   }
	   else
	   {
	      lt.tm_year = vcal.dtend->year - 1900;
	      lt.tm_mon = vcal.dtend->mon - 1;
	      lt.tm_mday = vcal.dtend->day;
//	      lt.tm_hour = vcal.dtend->hour + dif;
	      lt.tm_hour = vcal.dtend->hour;
	      lt.tm_min = vcal.dtend->min;
	      lt.tm_sec = vcal.dtend->sec;
	      dtend = mktime (&lt);

	      if (vcal.appointment_type == 1 || vcal.appointment_type == 2)
	         sprintf (hv0, "DTEND:%04d%02d%02dT235959Z\r\n",
	      	   vcal.dtend->year, vcal.dtend->mon, vcal.dtend->day);
	      else
	         sprintf (hv0, "DTEND:%04d%02d%02dT%02d%02d%02dZ\r\n",
	   	   vcal.dtend->year, vcal.dtend->mon, vcal.dtend->day,
		   vcal.dtend->hour, vcal.dtend->min, vcal.dtend->sec);
//		   vcal.dtend->hour + dif, vcal.dtend->min, vcal.dtend->sec);
	   }

	   strcat (Data, hv0);
	}

	if (!vcal.dtend && vcal.dend)
	{
	   lt.tm_year = vcal.dend->year - 1900;
	   lt.tm_mon = vcal.dend->mon - 1;
	   lt.tm_mday = vcal.dend->day;
//	   lt.tm_hour = vcal.dend->hour + dif;
	   lt.tm_hour = vcal.dend->hour;
	   lt.tm_min = vcal.dend->min;
	   lt.tm_sec = vcal.dend->sec;
	   dtend = mktime (&lt);

	   sprintf (hv0, "DTEND:%04d%02d%02dT235959Z\r\n",
	      	vcal.dend->year, vcal.dend->mon, vcal.dend->day);

	   strcat (Data, hv0);
	}

	/*
	 * Handelt es sich um einen sich wiederholenden Termin, muessen
	 * wir als erstes feststellen, in welchem Abstand sich der Termin
	 * wiederholt. Lotus Notes ist in diesem Punkt leider etwas
	 * umstaendlich.
	 */
	if (vcal.repeats && vcal.chstart_anz > 1)
	{
	struct _vcalendar_chaindt *dt1, *dt2;
	struct tm dt, *zeit;
	time_t t1, t2;
	double tdiff;

	   dt1 = vcal.chstart;
	   dt2 = vcal.chstart->next;
	   memset (&dt, 0, sizeof (time_t));

	   dt.tm_mday = dt1->day;
	   dt.tm_mon = dt1->mon - 1;
	   dt.tm_year = dt1->year - 1900;
	   dt.tm_hour = dt1->hour;
	   dt.tm_min = dt1->min;
	   dt.tm_sec = dt1->sec;

	   t1 = mktime (&dt);

	   dt.tm_mday = dt2->day;
	   dt.tm_mon = dt2->mon - 1;
	   dt.tm_year = dt2->year - 1900;
	   dt.tm_hour = dt2->hour;
	   dt.tm_min = dt2->min;
	   dt.tm_sec = dt2->sec;

	   t2 = mktime (&dt);
	   tdiff = difftime (t1, t2);

	   if (tdiff < 0.0)
	      tdiff /= -1;

	   /*
	    * 3600 = 1 hour
	    * 86400 = 1 day
	    * 604800 = 1 week
	    * 2419200 ~ 1 month
	    * 29030400 ~ 1 year
	    */
/*	   if (tdiff >= 86300 && tdiff < 86500)
	   {
	      strcat (Data, "RRULE:D1");
	      *
	       * Ist das wirklich ein taeglicher Termin oder
	       * lediglich eine mehrtaegige Veranstaltung?
	       *
	      if (vcal.chstart_anz < 7)
	      {
	         sprintf (hv0, " #%d\r\n", vcal.chstart_anz);
		 strcat (Data, hv0);
	      }
	      else
	         strcat (Data, "\r\n");
	   }
	   else */ if (tdiff >= 604700 && tdiff < 604900)
	   {
	      strcat (Data, "RRULE:W1 ");
	      /*
	       * Wir muessen noch den Wochentag feststellen...
	       */
	      zeit = localtime (&t1);

	      switch (zeit->tm_wday)
	      {
	         case 0: strcat (Data, "SO\r\n"); break;
		 case 1: strcat (Data, "MO\r\n"); break;
		 case 2: strcat (Data, "TU\r\n"); break;
		 case 3: strcat (Data, "WE\r\n"); break;
		 case 4: strcat (Data, "TH\r\n"); break;
		 case 5: strcat (Data, "FR\r\n"); break;
		 case 6: strcat (Data, "SA\r\n"); break;
	      }
	   }
	   else if (tdiff >= 2300000 && tdiff < 2500000)
	   {
	      sprintf (hv0, "RRULE:MD1 %d\r\n", dt1->day);
	      strcat (Data, hv0);
	   }
	   else if (tdiff >= 28000000 && tdiff < 30000000)
	   {
	      /*
	       * Wir muessen zuerst den Tag im Jahr finden ...
	       */
	      zeit = localtime (&t1);
	      sprintf (hv0, "RRULE:YD1 %d\r\n", zeit->tm_yday);
	      strcat (Data, hv0);
	   }
	}

	if (vcal.summary)
	{
	   strncpy (hv0, vcal.summary, 30);
	   hv0[30] = 0;
	   p = UnixToAscii (hv0);
	   sprintf (hv0, "DESCRIPTION:%s\r\n", p);
	   free (p);
	   strcat (Data, hv0);
	}

	/*
	 * Da Lotus Notes leider einige Eintraege mehrfach
	 * exportiert, muessen wir uns die Eckdaten eines
	 * Eintrags speichern. Am Beginn dieser function
	 * wird nach einem Eintrag gesucht und der erneute
	 * Eintrag abgelehnt falls er bereits existiert.
	 */
	AddSlot (dtstart, dtend, vcal.summary);
	strcat (Data, "END:VEVENT\r\nEND:VCALENDAR\r\n");
	repeat = 1;

	while ((i = SendData (fd, SlotNum, Data)) < 0 && repeat)
	{
	   if (repeat > 3)
	   {
	      lfprintf (10031, stderr, "Error %d while sending an appointment entry.\n", i);
	      break;
	   }

	   putchar ('.');
	   repeat++;
	   usleep (100);
	}

	if (Verbose)
	   lprintf (10032, "Send data to slot %d.\n", SlotNum);

	SlotNum++;
}

void FreeVEvent ()
{
	if (vcal.created)
	   free (vcal.created);

	if (vcal.UID)
	   free (vcal.UID);

	if (vcal.last_modified)
	   free (vcal.last_modified);

	if (vcal.dtstamp)
	   free (vcal.dtstamp);

	if (vcal.organizer)
	   free (vcal.organizer);

	if (vcal.vcal_att)
	{
	struct _vcalendar_attendee *akt, *prev;

	   akt = vcal.vcal_att;

	   while (akt)
	   {
	      if (akt->cn)
		 free (akt->cn);

	      if (akt->mailto)
		 free (akt->mailto);

	      prev = akt;
	      akt = akt->next;
	      free (prev);
	   }
	}

	if (vcal.chstart)
	{
	struct _vcalendar_chaindt *akt, *prev;

	   akt = vcal.chstart;

	   while (akt)
	   {
	      prev = akt;
	      akt = akt->next;
	      free (prev);
	   }
	}

	if (vcal.chend)
	{
	struct _vcalendar_chaindt *akt, *prev;

	   akt = vcal.chend;

	   while (akt)
	   {
	      prev = akt;
	      akt = akt->next;
	      free (prev);
	   }
	}

	if (vcal.description)
	   free (vcal.description);

	if (vcal.summary)
	   free (vcal.summary);

	if (vcal.categories)
	   free (vcal.categories);

	if (vcal.duration)
	   free (vcal.duration);

	if (vcal.dtstart)
	   free (vcal.dtstart);

	if (vcal.dtend)
	   free (vcal.dtend);

	if (vcal.vcal_valarm)
	{
	   if (vcal.vcal_valarm->vcal_trigger)
	   {
	   struct _vcalendar_trigger *tr;

	      tr = vcal.vcal_valarm->vcal_trigger;

	      if (tr->value)
		 free (tr->value);

	      if (tr->duration)
		 free (tr->duration);

	      free (tr);
	   }

	   free (vcal.vcal_valarm);
	}

	if (vcal.type)
	   free (vcal.type);

	if (vcal.location)
	   free (vcal.location);

	if (vcal.dstart)
	   free (vcal.dstart);

	if (vcal.dend)
	   free (vcal.dend);

	if (vcal.tzstart)
	{
	   if (vcal.tzstart->dlocal)
	      free (vcal.tzstart->dlocal);

	   if (vcal.tzstart->zn)
	      free (vcal.tzstart->zn);

	   free (vcal.tzstart);
	}

	if (vcal.tzend)
	{
	   if (vcal.tzend->dlocal)
	      free (vcal.tzend->dlocal);

	   if (vcal.tzend->zn)
	      free (vcal.tzend->zn);

	   free (vcal.tzend);
	}

	memset (&vcal, 0, sizeof (vcal));
	vcal.x_pilotstat = 1;
}

char *GetNotesAddress (char *str)
{
char *p;
int flag = 0;

	if (!str || !*str)
	   return NULL;

	if ((p = strchr (str, '@')) != NULL)
	   *p = 0;

	p = str;

	while ((p = strchr (p, '/')) != NULL)
	{
	   if (!flag)
	   {
	      *p = '@';
	      flag = 1;
	   }
	   else
	      *p = '.';

	   p++;
	}

	if ((p = strchr (str, ' ')) != NULL)
	   *p = '.';

	return str;
}

struct _X400 *GetX400Address (char *str)
{
char *sadr, *p, *np;
struct _X400 *x400;
int len;

	if (!str || !*str || strncasecmp (str, "CN=", 3))
	   return NULL;

	if ((sadr = MyStrdup (str)) == NULL)
	   return NULL;

	if ((x400 = (struct _X400 *)malloc (sizeof (struct _X400))) == NULL)
	{
	   lfprintf (10024, stderr, "Not enough memory!\n");
	   free (sadr);
	   return NULL;
	}

	memset (x400, 0, sizeof (struct _X400));
	p = sadr;

	if ((np = strchr (p, '/')) != NULL)
	   *np = 0;
	else
	   np = p;

	while (p)
	{
	   if (!strncasecmp (p, "CN=", 3))
	      x400->cn = MyStrdup (p+3);
	   else if (!strncasecmp (p, "OU=", 3))
	      x400->ou = MyStrdup (p+3);
	   else if (!strncasecmp (p, "O=", 2))
	      x400->o = MyStrdup (p+2);
	   else if (!strncasecmp (p, "C=", 2))
	      x400->c = MyStrdup (p+2);

	   if (p == np)
	      break;

	   np++;
	   p = np;

	   if (p && *p && (np = strchr (p, '/')) != NULL)
	      *np = 0;
	   else
	      np = p;
	}

	len = 0;

	if (x400->cn)
	   len += strlen (x400->cn);

	if (x400->ou)
	   len += strlen (x400->ou);

	if (x400->o)
	   len += strlen (x400->o);

	if (x400->c)
	   len += strlen (x400->c);

	if (!len)		// Should not happen, but ...
	{
	   if (x400->cn)
	      free (x400->cn);

	   if (x400->ou)
	      free (x400->ou);

	   if (x400->o)
	      free (x400->o);

	   if (x400->c)
	      free (x400->c);

	   free (x400);
	   free (sadr);
	   return NULL;
	}

	if ((x400->inet = (char *)malloc (len+8)) == NULL)
	{
	   if (x400->cn)
	      free (x400->cn);

	   if (x400->ou)
	      free (x400->ou);

	   if (x400->o)
	      free (x400->o);

	   if (x400->c)
	      free (x400->c);

	   free (x400);
	   free (sadr);
	   lfprintf (10024, stderr, "Not enough memory!\n");
	   return NULL;
	}

	*x400->inet = 0;

	if (x400->cn)
	{
	   if ((p = strchr (x400->cn, ' ')) != NULL)
	      *p = '.';

	   strcat (x400->inet, x400->cn);

	   if (p)
	      *p = ' ';
	}

	if (x400->ou)
	{
	   strcat (x400->inet, "@");
	   strcat (x400->inet, x400->ou);
	}

	if (x400->o)
	{
	   strcat (x400->inet, ".");
	   strcat (x400->inet, x400->o);
	}

	if (x400->c)
	{
	   strcat (x400->inet, ".");
	   strcat (x400->inet, x400->c);
	}

	free (sadr);
	return x400;
}

struct _vcalendar_datetime *ConvertDateTime (char *ret, char *str)
{
int flag = 0;
int uhrzeit = 0;
int datum = 0;
char *p, *pos, *ln, part[5];
struct _vcalendar_datetime *dt;

	if (!str && (strlen (str) != 8 || strlen (str) != 10 || strlen (str) != 19))
	{
	   lfprintf (10080, stderr, "Non or illegal date detected!\n");
	   return NULL;
	}

	if ((ln = MyStrdup (str)) == NULL)
	   return NULL;

	pos = ln;

	if (ret)
	   *ret = 0;

	if ((dt = (struct _vcalendar_datetime *)malloc (sizeof (struct _vcalendar_datetime))) == NULL)
	{
	   lfprintf (10032, stderr, "Not enough memory for a date segment!\n");
	   return NULL;
	}

	memset (dt, 0, sizeof (struct _vcalendar_datetime));

	if ((p = strchr (pos, '.')) == NULL)
	{
	   if ((p = strchr (pos, '/')) == NULL)
	   {
#ifdef DEBUG
	      fprintf (stderr, "ConvertDateTime: Uhrzeit gefunden >%s<\n", pos);
#endif
	      uhrzeit = 1;
	   }

	   flag = 1;
	}

	if (!uhrzeit)
	{
	   strncpy (part, p-2, 2);
	   part[2] = 0;

	   if (flag)			// US Date: day and month changed
	      dt->mon = atoi (part);
	   else
	      dt->day = atoi (part);

	   pos = p + 1;
	   flag = 0;

	   if ((p = strchr (pos, '.')) == NULL)
	   {
	      if ((p = strchr (pos, '/')) == NULL)
	      {
	         free (dt);
		 free (ln);
	         return NULL;
	      }

	      flag = 1;
	   }

	   strncpy (part, p-2, 2);
	   part[2] = 0;

	   if (flag)			// US Date: day and month changed
	      dt->day = atoi (part);
	   else
	      dt->mon = atoi (part);

	   pos = p + 1;

	   if ((p = strchr (pos, ' ')) == NULL)		// date only?
	   {
	      datum = 1;
	      p = pos + 4;
	   }

	   strncpy (part, p-4, 4);
	   part[4] = 0;
	   dt->year = atoi (part);

	   if (datum)
	   {
	      free (ln);
	      return dt;
	   }

	   pos = p + 1;
	}

	/*
	 * Der String enthaelt auch, oder nur eine Uhrzeit. Diese besteht
	 * immer aus Stunden, Minuten und Sekunden, jeweils durch einen
	 * Doppelpunkt getrennt.
	 */
	if ((p = strchr (pos, ':')) == NULL)
	{
	   free (dt);
	   free (ln);
	   return NULL;
	}

	strncpy (part, p-2, 2);
	part[2] = 0;
	dt->hour = atoi (part);
	pos = p + 1;

	if ((p = strchr (pos, ':')) == NULL)
	{
	   free (dt);
	   free (ln);
	   return NULL;
	}

	strncpy (part, p-2, 2);
	part[2] = 0;
	dt->min = atoi (part);
	p++;
	strncpy (part, p, 2);
	part[2] = 0;
	dt->sec = atoi (part);
	p += 2;

	if (*p && *(p+1) && *p == ' ' && *(p+1) == 'P' && *(p+2) == 'M')
	{
	   if (dt->hour != 12)
	      dt->hour += 12;
	}

//	if (!uhrzeit)
//	   TimeZone (dt, 0);

	if (ret)
	   sprintf (ret, "%04d%02d%02dT%02d%02d%02dZ", dt->year, dt->mon,
	   		dt->day, dt->hour, dt->min, dt->sec);

	free (ln);
	return dt;
}

int TimeZone (struct _vcalendar_datetime *date, int zeit)
{
time_t t;
struct tm lt, *tp;
long dif;
struct _vcalendar_timezone *tz;

	if (!date)
	   return false;

	memset (&lt, 0, sizeof (struct tm));
	lt.tm_year = date->year - 1900;
	lt.tm_mon = date->mon - 1;
	lt.tm_mday = date->day;
	lt.tm_hour = date->hour;
	lt.tm_min = date->min;
	lt.tm_sec = date->sec;

	if (vcal.tzstart)
	   tz = vcal.tzstart;
	else if (vcal.tzend)
	   tz = vcal.tzend;
	else
	   tz = NULL;

	if (tz)
	   lt.tm_isdst = tz->doffset;
	else
	   lt.tm_isdst = -1;

	t = mktime (&lt);

	if (ZeitDifferenz)
	{
	   dif = (ZeitDifferenz + 1) * 3600L;		// Zeitdifferenz in Sekunden
	   t += dif;
	   tp = gmtime (&t);

	   if (!zeit)
	   {
	      date->year = tp->tm_year + 1900;
	      date->mon = tp->tm_mon + 1;
	      date->day = tp->tm_mday;
	   }

	   date->hour = tp->tm_hour;
	   date->min = tp->tm_min;
	   date->sec = tp->tm_sec;
	}
	else if (!LocalTime)
	{
	   // Zeit in UTC umwandeln
	   tp = gmtime (&t);

	   if (!zeit)
	   {
	      date->year = tp->tm_year + 1900;
	      date->mon = tp->tm_mon + 1;
	      date->day = tp->tm_mday;
	   }

	   date->hour = tp->tm_hour;
	   date->min = tp->tm_min;
	   date->sec = tp->tm_sec;
	}

	return true;
}

struct _vcalendar_timezone *GetTimeZone (char *zone)
{
char *p, *ln, *np;
struct _vcalendar_timezone *tz;

	if (!zone || !*zone)
	   return NULL;

	if ((tz = (struct _vcalendar_timezone *)malloc (sizeof (struct _vcalendar_timezone))) == NULL)
	{
	   lfprintf (10083, stderr, "Not enough memory for saving the time zone information!\n");
	   return NULL;
	}

	memset (tz, 0, sizeof (struct _vcalendar_timezone));

	if ((ln = MyStrdup (zone)) == NULL)
	{
	   free (tz);
	   return NULL;
	}

	/*
	 * Die Zeitzoneninformation kann wie folgt aussehen:
	 * Z=-1$DO=1$DL=3 -1 1 10 -1 1$ZX=70$ZN=W. Europe
	 */
	p = ln;

	if ((np = strchr (p, '$')) != NULL)
	   *np = 0;
	else
	   np = p;

	while (p)
	{
	   if (!strncasecmp (p, "Z=", 2))
	      tz->base = atoi (p+2);
	   else if (!strncasecmp (p, "DO=", 3))
	      tz->doffset = atoi (p+3);
	   else if (!strncasecmp (p, "DL=", 3))
	   {
	      if ((tz->dlocal = MyStrdup (p+3)) == NULL)
	      {
	         if (tz->zn)
		    free (tz->zn);

		 free (tz);
		 free (ln);
		 return NULL;
	      }
	   }
	   else if (!strncasecmp (p, "ZX=", 3))
	      tz->zx = atoi (p+3);
	   else if (!strncasecmp (p, "ZN=", 3))
	   {
	      if ((tz->zn = MyStrdup (p+3)) == NULL)
	      {
	         if (tz->dlocal)
		    free (tz->dlocal);

		 free (tz);
		 free (ln);
		 return NULL;
	      }
	   }

	   if (np == p)
	      break;

	   np++;
	   p = np;

	   if ((np = strchr (p, '$')) != NULL)
	      *np = 0;
	   else
	      np = p;
	}

	free (ln);
	return tz;
}

int AddSlot (time_t start, time_t end, char *summary)
{
struct _slot *akt, *sl;

	if (slot == NULL)
	{
	   if ((slot = (struct _slot *)malloc (sizeof (struct _slot))) == NULL)
	   {
	      lfprintf (10024, stderr, "Not enough memory!\n");
	      return -1;
	   }

	   memset (slot, 0, sizeof (struct _slot));
	   slot->start = start;
	   slot->end = end;
	   slot->summary = strdup (summary);
	   return 0;
	}

	akt = slot;

	while (akt)
	{
	   sl = akt;
	   akt = akt->next;
	}

	if ((akt = (struct _slot *)malloc (sizeof (struct _slot))) == NULL)
	{
	   lfprintf (10024, stderr, "Not enough memory!\n");
	   return -1;
	}

	memset (akt, 0, sizeof (struct _slot));
	akt->start = start;
	akt->end = end;
	akt->summary = strdup (summary);
	sl->next = akt;
	return 0;
}

struct _slot *FindSlot (time_t start, time_t end, char *summary)
{
struct _slot *akt;

	akt = slot;

	while (akt)
	{
	   if (akt->start == start && akt->end == end && !strcmp (akt->summary, summary))
	      return akt;

	   akt = akt->next;
	}

	return NULL;
}

struct _slot *FindSlotSummary (char *summary)
{
struct _slot *akt;

	akt = slot;

	while (akt)
	{
	   if (!strcmp (akt->summary, summary))
	      return akt;

	   akt = akt->next;
	}

	return NULL;
}

void FreeSlot ()
{
struct _slot *akt, *sl;

	akt = slot;

	while (akt)
	{
	   sl = akt;
	   akt = akt->next;
	   free (sl->summary);
	   free (sl);
	}
}

char *WinToUnix (char *str)
{
int len, i;
char *buf, *p, *pos;

	len = strlen (str);
	i = 0;

	if (len <= 0)
	   return NULL;

	if ((buf = (char *)malloc (len * 2 + 5)) == NULL)
	{
	   lfprintf (10034, stderr, "Not enough memory to convert special characters!\n");
	   return NULL;
	}

	memset (buf, 0, len * 2 + 5);
	p = buf;
	pos = str;

	while (i < len)
	{
	   switch (*pos)
	   {
	      case '':
	         *p++ = '\xc3';
		 *p = '\xa4';
	      break;

	      case '':
	         *p++ = '\xc3';
		 *p = '\xb6';
	      break;

	      case '':
	         *p++ = '\xc3';
		 *p = '\xbc';
	      break;

	      case '':
	         *p++ = '\xc3';
		 *p = '\x9f';
	      break;

	      case '':
	         *p++ = '\xc3';
		 *p = '\x84';
	      break;

	      case '':
	         *p++ = '\xc3';
		 *p = '\x96';
	      break;

	      case '':
	         *p++ = '\xc3';
		 *p = '\x9c';
	      break;

	      case '':
	         *p++ = '\xc2';
		 *p = '\xa7';
	      break;

	      case ',':
	         *p++ = '\\';
		 *p = ',';
	      break;

	      case ';':
	         *p++ = '\\';
		 *p = ';';
	      break;

	      case '\n':
		 *p++ = '\\';
		 *p++ = 'n';
	         *p++ = *pos;
		 *p = ' ';
	      break;

	      case '\r':
	         p--;
	      break;

	      default:
	         *p = *pos;
	   }

	   pos++;
	   p++;
	   i++;
	}

	return buf;
}

char *UnixToAscii (char *str)
{
int len, i;
char *buf, *p, *pos;

	len = strlen (str);
	i = 0;

	if (len <= 0)
	   return NULL;

	if ((buf = (char *)malloc (len * 2 + 5)) == NULL)
	{
	   lfprintf (10034, stderr, "Not enough memory to convert special characters!\n");
	   return NULL;
	}

	memset (buf, 0, len * 2 + 5);
	p = buf;
	pos = str;

	while (i < len)
	{
	   if (*pos == '\xc3')
	   {
	      pos++;

	      switch (*pos)
	      {
	         case '\xa4':
		    *p++ = 'a';
		    *p = 'e';
	         break;

	         case '\xb6':
	            *p++ = 'o';
		    *p = 'e';
	         break;

	         case '\xbc':
	            *p++ = 'u';
		    *p = 'e';
	         break;

	         case '\x9f':
	            *p++ = 's';
		    *p = 's';
	         break;

	         case '\x84':
	            *p++ = 'A';
		    *p = 'E';
	         break;

	         case '\x96':
	            *p++ = 'O';
		    *p = 'E';
	         break;

	         case '\x9c':
	            *p++ = 'U';
		    *p = 'E';
	         break;

	         default:
	            *p = *pos;
	      }
	   }
	   else
	      *p = *pos;

	   pos++;
	   p++;
	   i++;
	}

	return buf;
}

char *GetLine (char *pos)
{
char *p1;

	if (!pos || *pos == 0)
	   return NULL;

	p1 = pos;

	while (*p1 && *p1 != 0x0d && *p1 != 0x0a)
	   p1++;

	if (*p1 == 0x0d || *p1 == 0x0a)
	{
	   if (*(p1+1) == 0x0d || *(p1+1) == 0x0a)
	   {
	      *p1 = *(p1+1) = 0;
	      p1 += 2;

	   }
	   else
	   {
	      *p1 = 0;
	      p1++;
	   }
	}

	return p1;
}

char *AvoidNull (char *str, int len)
{
int j, i = 0;
char *p;

	if (len <= 0)
	   return str;

	p = str;
	j = 0;

	while (i < len)
	{
	   if (*p == 0)
	   {
	      if (Verbose)
	         lprintf (10035, "%d. Found a NULL at position %d.\n", ++j, i);

	      *p = ' ';
	   }

	   i++;
	   p++;
	}

	*(str+len) = 0;
	return str;
}

char *rtrim (char *str)
{
char *p;

	p = str;

	while (*p && *p != ' ')
	   p++;

	*p = 0;
	return str;
}

char *ltrim (char *str)
{
char *p1, *p2;

	p1 = p2 = str;

	while (*p1 && *p1 == ' ')
	   p1++;

	while (*p1)
	{
	   *p2 = *p1;
	   p1++;
	   p2++;
	}

	*p2 = 0;
	return str;
}

char *MyStrdup (char *str)
{
char *p;
int len;

	if (!str || !*str)
	   return NULL;

	len = strlen (str);

	if ((p = (char *)calloc (len+1, sizeof (char))) == NULL)
	{
	   lfprintf (10024, stderr, "Not enough memory!\n");
	   return NULL;
	}

	strcpy (p, str);
	return p;
}

void ShowHelp ()
{
	lprintf (10036, "usage: handy -a <file> [-t [<+/- hours]] [-o <file>] [-w] [-d <device>]\n");
	lprintf (10037, "             [-p <pin>] [-b <baud>]\n");
	lprintf (10038, "       handy -h\n\n");
	lprintf (10039, "-a <file>            Exported file out of Lotus Notes,\n");
	lprintf (10040, "--appointment-file<f>that contains all appointments.\n\n");
	lprintf (10041, "-o <file>            To vCal Version 2.0 compatible file, that is\n");
	lprintf (10042, "--outfile <file>     known, for example, by Organizer of KDE 3.x.\n");
	lprintf (10043, "                     If this parameter is omitted, it creates a file\n");
	lprintf (10044, "                     called \"icalout.ics\" in the actual directory.\n");
	lprintf (10045, "                     If this file already exist, it is truncated\n");
	lprintf (10046, "                     and overwritten!\n\n");
	lprintf (10047, "-t [+/-hours]        If this parameter is omitted, all clock times\n");
	lprintf (10048, "--local-time [+/-ho] remain unchanged. Otherwise the specified clock\n");
	lprintf (10049, "                     times are added or subtracted.\n\n");
	lprintf (10050, "-w                   The appointments are send to a connected device.\n");
	lprintf (10051, "--write-cell-phone\n\n");
	lprintf (10052, "-d <device>          The port your device (phone) is connected to.\n");
	lprintf (10053, "--device <device>    If this parameter is omitted, but the parameter");
	lprintf (10054, "                     -w is not, the port /dev/ttyS0 (COM1:) is used.\n");
	lprintf (10055, "                     \n");
	lprintf (10056, "-b <speed>           This parameter sets the baoud rate of the serial\n");
	lprintf (10057, "--baud <speed>       device. The values 19200, 38400, 57600 and 115200\n");
	lprintf (10058, "                     are allowed. If this parameter is omitted, the\n");
	lprintf (10059, "                     speed is set to 19200. Make sure that your\n");
	lprintf (10060, "                     connected device is set the same speed.\n\n");
	lprintf (10061, "-p <pin>             If your phone is protected witha PIN, you must\n");
	lprintf (10062, "--pin <pin>          set this PIN with this parameter.\n");
	lprintf (10063, "                     \n");
	lprintf (10064, "-v                   Verbose; The program outputs a lot of information.\n");
	lprintf (10065, "--verbose            This information can be used for error detection.\n");
	lprintf (10066, "                     \n");
	lprintf (10067, "-h / --help          Shows this help.\n\n");
}

void ShowVersion ()
{
	printf ("Handy Version %d.%d.%d\n", VMAJOR, VMINOR, VCOUNT);
	printf ("Author: Andreas Theofilu (Theos Soft) <andreas@TheosSoft.net>\n\n");
}

